from collections.abc import Iterable
from typing import Any, Callable

import lightgbm as lgb
import numpy as np
from sklearn.model_selection import BaseCrossValidator

from imlightgbm.docstring import add_docstring
from imlightgbm.objective import set_params


@add_docstring("train")
def train(
    params: dict[str, Any],
    train_set: lgb.Dataset,
    num_boost_round: int = 100,
    valid_sets: list[lgb.Dataset] = None,
    valid_names: list[str] = None,
    init_model: str | lgb.Path | lgb.Booster | None = None,
    keep_training_booster: bool = False,
    callbacks: list[Callable] | None = None,
) -> lgb.Booster:
    _params = set_params(params=params, train_set=train_set)
    return lgb.train(
        params=_params,
        train_set=train_set,
        num_boost_round=num_boost_round,
        valid_sets=valid_sets,
        valid_names=valid_names,
        init_model=init_model,
        keep_training_booster=keep_training_booster,
        callbacks=callbacks,
    )


@add_docstring("cv")
def cv(
    params: dict[str, Any],
    train_set: lgb.Dataset,
    num_boost_round: int = 100,
    folds: Iterable[tuple[np.ndarray, np.ndarray]] | BaseCrossValidator | None = None,
    nfold: int = 5,
    stratified: bool = True,
    shuffle: bool = True,
    init_model: str | lgb.Path | lgb.Booster | None = None,
    fpreproc: Callable[
        [lgb.Dataset, lgb.Dataset, dict[str, Any]],
        tuple[lgb.Dataset, lgb.Dataset, dict[str, Any]],
    ]
    | None = None,
    seed: int = 0,
    callbacks: list[Callable] | None = None,
    eval_train_metric: bool = False,
    return_cvbooster: bool = False,
) -> dict[str, list[float] | lgb.CVBooster]:
    _params = set_params(params=params, train_set=train_set)
    return lgb.cv(
        params=_params,
        train_set=train_set,
        num_boost_round=num_boost_round,
        folds=folds,
        nfold=nfold,
        stratified=stratified,
        shuffle=shuffle,
        init_model=init_model,
        fpreproc=fpreproc,
        seed=seed,
        callbacks=callbacks,
        eval_train_metric=eval_train_metric,
        return_cvbooster=return_cvbooster,
    )



from typing import Callable

_space = "\n        "
ALL_PARAMS = {
    "params": f"dict{_space}Parameters for training. Values passed through ``params`` take precedence over those supplied via arguments.",
    "train_set": f"Dataset{_space}Data to be trained on.",
    "num_boost_round": f"int, optional (default=100){_space}Number of boosting iterations.",
    "valid_sets": f"list of Dataset, or None, optional (default=None){_space}List of data to be evaluated on during training.",
    "valid_names": f"list of str, or None, optional (default=None){_space}Names of ``valid_sets``.",
    "folds": f"generator or iterator of (train_idx, test_idx) tuples, scikit-learn splitter object or None, optional (default=None){_space}If generator or iterator, it should yield the train and test indices for each fold.{_space}If object, it should be one of the scikit-learn splitter classes{_space}(https://scikit-learn.org/stable/modules/classes.html#splitter-classes){_space}and have ``split`` method.{_space}This argument has highest priority over other data split arguments.",
    "nfold": f"int, optional (default=5){_space}Number of folds in CV.",
    "stratified": f"bool, optional (default=True){_space}Whether to perform stratified sampling.",
    "shuffle": f"bool, optional (default=True){_space}Whether to shuffle before splitting data.",
    "init_model": f"str, pathlib.Path, Booster or None, optional (default=None){_space}Filename of LightGBM model or Booster instance used for continue training.",
    "fpreproc": f"callable or None, optional (default=None){_space}Preprocessing function that takes (dtrain, dtest, params) and returns transformed versions of those.",
    "seed": f"int, optional (default=0){_space}Seed used to generate the folds (passed to numpy.random.seed).",
    "keep_training_booster": f"bool, optional (default=False){_space}Whether the returned Booster will be used to keep training.{_space}If False, the returned value will be converted into _InnerPredictor before returning.{_space}This means you won't be able to use ``eval``, ``eval_train`` or ``eval_valid`` methods of the returned Booster.{_space}When your model is very large and cause the memory error,{_space}you can try to set this param to ``True`` to avoid the model conversion performed during the internal call of ``model_to_string``.{_space}You can still use _InnerPredictor as ``init_model`` for future continue training.",
    "callbacks": f"list of callable, or None, optional (default=None){_space}List of callback functions that are applied at each iteration.{_space}See Callbacks in Python API for more information.",
    "eval_train_metric": f"bool, optional (default=False){_space}Whether to display the train metric in progress.",
    "return_cvbooster": f"bool, optional (default=False){_space}Whether to return Booster models trained on each fold through ``CVBooster``.",
}


PARAMS_MAPPER = {
    "train": {
        "description": "Perform the training with given parameters.",
        "selected_params": [
            "params",
            "train_set",
            "num_boost_round",
            "valid_sets",
            "valid_names",
            "init_model",
            "keep_training_booster",
            "callbacks",
        ],
        "return_description": f"booster: Booster{_space}The trained Booster model.",
    },
    "cv": {
        "description": "Perform the cross-validation with given parameters.",
        "selected_params": [
            "params",
            "train_set",
            "num_boost_round",
            "folds",
            "nfold",
            "stratified",
            "shuffle",
            "init_model",
            "fpreproc",
            "seed",
            "callbacks",
            "eval_train_metric",
            "return_cvbooster",
        ],
        "return_description": "eval_results: dict\n        History of evaluation results of each metric.\n        The dictionary has the following format:\n        {'valid metric1-mean': [values], 'valid metric1-stdv': [values],\n        'valid metric2-mean': [values], 'valid metric2-stdv': [values],\n        ...}.\n        If ``return_cvbooster=True``, also returns trained boosters wrapped in a ``CVBooster`` object via ``cvbooster`` key.\n        If ``eval_train_metric=True``, also returns the train metric history.\n        In this case, the dictionary has the following format:\n        {'train metric1-mean': [values], 'valid metric1-mean': [values],\n        'train metric2-mean': [values], 'valid metric2-mean': [values],\n        ...}.",
    },
}


def generate_docstring(
    description: str,
    selected_params: list[str],
    return_description: str = "",
) -> str:
    """Generate a docstring with a provided description, selected parameters, and optional return description."""
    docstring = f"{description}\n\n    Parameters\n    ----------\n"
    for param in selected_params:
        docstring += f"    {param}: {ALL_PARAMS[param]}\n"
    if return_description:
        docstring += f"\n    Returns\n    -------\n    {return_description}\n"
    return docstring


def add_docstring(func_name: str) -> Callable:
    """Decorator to add a docstring to a function based on provided parameters and descriptions."""

    def decorator(func: Callable) -> Callable:
        func.__doc__ = generate_docstring(**PARAMS_MAPPER[func_name])
        return func

    return decorator



from copy import deepcopy
from functools import partial
from typing import Any, Callable

import numpy as np
from lightgbm import Dataset
from sklearn.utils.multiclass import type_of_target

from imlightgbm.base import Metric, Objective, SupportedTask

ObjLike = Callable[[np.ndarray, Dataset], tuple[np.ndarray, np.ndarray]]
ALPHA_DEFAULT: float = 0.25
GAMMA_DEFAULT: float = 2.0
OBJECTIVE_STR: str = "objective"
METRIC_STR: str = "metric"
IS_HIGHER_BETTER: bool = False


def _power(num_base: np.ndarray, num_pow: float):
    """Safe power."""
    return np.sign(num_base) * (np.abs(num_base)) ** (num_pow)


def _log(array: np.ndarray, is_prob: bool = False) -> np.ndarray:
    """Safe log."""
    _upper = 1 if is_prob else None
    return np.log(np.clip(array, 1e-6, _upper))


def _sigmoid(x: np.ndarray) -> np.ndarray:
    """Convert raw predictions to probabilities in binary task."""
    return 1 / (1 + np.exp(-x))


def binary_focal_objective(
    pred: np.ndarray, train_data: Dataset, gamma: float
) -> tuple[np.ndarray, np.ndarray]:
    """Return grad, hess for binary focal objective."""
    label = train_data.get_label()
    pred_prob = _sigmoid(pred)

    # gradient
    g1 = pred_prob * (1 - pred_prob)
    g2 = label + ((-1) ** label) * pred_prob
    g3 = pred_prob + label - 1
    g4 = 1 - label - ((-1) ** label) * pred_prob
    g5 = label + ((-1) ** label) * pred_prob
    grad = gamma * g3 * _power(g2, gamma) * _log(g4) + ((-1) ** label) * _power(
        g5, (gamma + 1)
    )

    # hess
    h1 = _power(g2, gamma) + gamma * ((-1) ** label) * g3 * _power(g2, (gamma - 1))
    h2 = ((-1) ** label) * g3 * _power(g2, gamma) / g4
    hess = ((h1 * _log(g4) - h2) * gamma + (gamma + 1) * _power(g5, gamma)) * g1
    return grad, hess


def binary_weighted_objective(pred: np.ndarray, train_data: Dataset, alpha: float):
    """Return grad, hess for binary weighted objective."""
    label = train_data.get_label()
    pred_prob = _sigmoid(pred)
    grad = -(alpha**label) * (label - pred_prob)
    hess = (alpha**label) * pred_prob * (1.0 - pred_prob)
    return grad, hess


def multiclass_focal_objective(
    pred: np.ndarray, train_data: Dataset, alpha: float, gamma: float
) -> tuple[np.ndarray, np.ndarray]:
    # TODO
    return


def multiclass_weighted_objective(
    pred: np.ndarray, train_data: Dataset, alpha: float, gamma: float
) -> tuple[str, float, bool]:
    # TODO
    return


def _get_metric(task_enum: SupportedTask, metric: str | None) -> str:
    """Retrieve the appropriate metric function based on task."""
    metric_mapper: dict[SupportedTask, list[Metric]] = {
        SupportedTask.binary: [Metric.auc, Metric.binary_error, Metric.binary_logloss],
        SupportedTask.multiclass: [
            Metric.auc_mu,
            Metric.multi_logloss,
            Metric.multi_error,
        ],
    }
    if metric:
        metric_enum = Metric.get(metric)
        metric_enums = metric_mapper[task_enum]
        if metric_enum not in metric_enums:
            valid_metrics = ", ".join([m.value for m in metric_enums])
            raise ValueError(f"Invalid metric: Supported metrics are {valid_metrics}")
        return metric_enum.value

    return metric_mapper[task_enum][0].value


def _get_objective(
    task_enum: SupportedTask, objective: str | None, alpha: float, gamma: float
) -> ObjLike:
    """Retrieve the appropriate objective function based on task and objective type."""
    objective_mapper: dict[SupportedTask, dict[Objective, ObjLike]] = {
        SupportedTask.binary: {
            Objective.focal: partial(binary_focal_objective, gamma=gamma),
            Objective.weighted: partial(binary_weighted_objective, alpha=alpha),
        },
        SupportedTask.multiclass: {
            Objective.focal: partial(
                multiclass_focal_objective, alpha=alpha, gamma=gamma
            ),
            Objective.weighted: partial(
                multiclass_weighted_objective, alpha=alpha, gamma=gamma
            ),
        },
    }
    if objective:
        objective_enum = Objective.get(objective)
        return objective_mapper[task_enum][objective_enum]

    return objective_mapper[task_enum][Objective.focal]


def _get_fobj_feval(
    train_set: Dataset,
    alpha: float,
    gamma: float,
    objective: str | None,
    metric: str | None,
) -> tuple[ObjLike, str]:
    """Return obj and eval with respect to task type."""
    _task = type_of_target(train_set.get_label())
    task_enum = SupportedTask.get(_task)
    # FIXME: remove after developing multiclass objective
    if task_enum != SupportedTask.binary:
        raise ValueError(
            "Inferred task is not binary. Multiclass classification will be supported starting from version 0.1.0."
        )
    feval = _get_metric(task_enum=task_enum, metric=metric)
    fobj = _get_objective(
        task_enum=task_enum, objective=objective, alpha=alpha, gamma=gamma
    )
    return fobj, feval


def set_params(params: dict[str, Any], train_set: Dataset) -> dict[str, Any]:
    """Set params and eval finction, objective in params."""
    _params = deepcopy(params)
    _objective = _params.pop(OBJECTIVE_STR, None)
    _metric = _params.pop(METRIC_STR, None)

    if _metric and not isinstance(_metric, str):
        raise ValueError("metric must be str")

    fobj, feval = _get_fobj_feval(
        train_set=train_set,
        alpha=_params.pop("alpha", ALPHA_DEFAULT),
        gamma=_params.pop("gamma", GAMMA_DEFAULT),
        objective=_objective,
        metric=_metric,
    )
    _params.update({OBJECTIVE_STR: fobj, METRIC_STR: feval})
    return _params
